# -*- coding: utf-8 -*-
"""
Created on Wed Feb  9 13:02:33 2022

This module is the script use to train the Doc2Vec model that gets used
in SINCT

The Census Bureau has reviewed this data product to ensure appropriate 
access, use, and disclosure avoidance protection of the confidential 
source data used to produce this product (Data Management System (DMS)
number: P-7504847, subproject P-7514952, Disclosure Review Board 
(DRB) approval number: CBDRB-FY24-EWD001-003).


"""

import logging
import pandas as pd
import numpy as np
from nltk.corpus import stopwords
from nltk.stem import SnowballStemmer
import smart_open

smart_open.open = smart_open.smart_open

from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from gensim import utils


class NapcsModeler:
    """Napcs_modeler reads in the existing corpus of data and trains
    a Doc2Vec model on the Distributed Bag of Words (DBOW) task. This yields
    document and word embeddings that we can use in SINCT to query for
    similar training corpus examples to an incoming search.

    NapcsModeler is expected to run offline whenever we have new data
    to add to the training corpus. Since the real training data are Title 13
    protected, we have provided an alternative dataset based on the publicly
    available Amazon reviews dataset https://amazon-reviews-2023.github.io/
    
    Bridging Language and Items for Retrieval and Recommendation
       Yupeng Hou, Jiacheng Li, Zhankui He, An Yan, Xiusi Chen, Julian McAuley
       arXiv https://arxiv.org/pdf/2403.03952
       
    In this example dataset:
        1. title: replaces a user's search text in the NAPCS dataset
        2. main_category: takes the place of the NAICS description in the
           original training dataset
        3. parent_asin: takes the place of the NAPCS code in the original
           training dataset

    See Le & Mikolov (2014) for more information on Doc2Vec.

    """

    def __init__(self) -> None:
        # stop_words come from NLTK's list of stopwords
        self.stop_words = set(stopwords.words('english'))
        # SnowballStemmer is an improved version of the Porter Stemmer
        self.snowball = SnowballStemmer(language='english')
        # raw training corpus NAPCS examples
        self.examples = self.load_examples()
        self.y_labels = self.examples['parent_asin'].values
        np.save('./labels.npy', self.y_labels)

    def load_examples(self) -> pd.DataFrame:
        """load_examples function brings the training corpus of product
        examples into memory so we can train the Doc2Vec model on them.

        Args:
            self: the class

        Returns:
            pd.Dataframe: examples is all of the product data including code,
                title, and main category
        """

        examples = pd.read_csv('./meta_Video_Games.csv')
        examples = examples[['main_category', 'title', 'parent_asin']]
        examples['text'] = (
            examples['title'] + ' ' +
            examples['main_category']
        )

        return examples


    def stemmer(self, token: str) -> str:
        """
        stemmer takes a token and stems it
        using a Snowball Stemmer

        Parameter:
            token: a single string token
        Return:
            cleaned_token: the stemmed token
        """
        cleaned_token = self.snowball.stem(token)
        return cleaned_token

    def text_cleaner(self, search: str) -> list:
        """
        text_cleaner takes an input search and a naics_title and returns
        a list of tokens that have been stemmed. Stopwords
        and numbers have also been removed.

        Parameters:
            search: what the respondent typed into the search box

        Returns:
            tokens: a list of tokens without numbers and stop words that has
                    been stemmed
        """
        cleaned_tokens = utils.simple_preprocess(str(search))
        cleaned_tokens = [
            self.stemmer(token) for token in cleaned_tokens
            if (token not in self.stop_words and not token.isdigit())
        ]
        return cleaned_tokens

    def read_corpus(self, examples: pd.Series) -> TaggedDocument:
        """
        read_corpus takes a dataframe of example searches and yields a stream
        of gensim TaggedDocuments to use in training

        Parameter:
            examples: a pandas Series of the example searches
        Returns:
            TaggedDocument stream
        """
        for i, search in examples.iteritems():
            cleaned_tokens = self.text_cleaner(search)
            tagged_doc = TaggedDocument(cleaned_tokens, [i])
            yield tagged_doc

    def train(self, out_loc=None) -> None:
        """train model is where we actually train the Doc2Vec model with the
        following parameters:
            vector_size: the number of nodes in the hidden layer (60)
            window: the maximum number of surrounding words the model should
                    look at for context (5)
            dm: whether to use distributed memory (1) or distributed bow (0)
            min_count: the minimum number of times a word should appear to be
                       considered by the model (1)
            workers: how many cores the model should use to train on (6)
            epochs: the number of training steps (50)
            dbow_words: whether we should train word vectors in skip-gram
                        fashion first (1 = yes)

        Parameters:
            self: the class
            out_loc: a string representing an alternative filepath to save
                     the trained model
        """
        # set default model name depending on whether it's a NAICS model or not
        if out_loc is None:
            out_loc = './SINCT_model.model'

        # setting the logger to show training info in the log
        logging.basicConfig(
            format='%(asctime)s : %(levelname)s : %(message)s',
            level=logging.INFO)
        train_corpus = list(self.read_corpus(self.examples['text']))
        napcs_model = Doc2Vec(vector_size=60, window=5, dm=0, min_count=1,
                              workers=6, epochs=50, dbow_words=1)
        napcs_model.build_vocab(train_corpus)
        napcs_model.train(train_corpus,
                          total_examples=napcs_model.corpus_count,
                          epochs=napcs_model.epochs)
        napcs_model.save(out_loc)
