# -*- coding: utf-8 -*-
"""
Created on Fri Jan 28 15:03:42 2022

This module implements the SINCT API that a survey can call
in the Economic Census

The Census Bureau has reviewed this data product to ensure appropriate 
access, use, and disclosure avoidance protection of the confidential 
source data used to produce this product (Data Management System (DMS)
number: P-7504847, subproject P-7514952, Disclosure Review Board 
(DRB) approval number: CBDRB-FY24-EWD001-003).


SINCT stands for Smart Instrument NAPCS Classification Tool
and is used by respondents to query for NAPCS products in the
Economic Census. Users can provide a text write-in of what their
products/services are and SINCT will return up to 10 recommended
NAPCS codes for their business. It is based on a Doc2Vec model.

Since the original training data are Title 13 protected, this uses an
example dataset from the publicly available Amazon reviews dataset
https://amazon-reviews-2023.github.io/

Bridging Language and Items for Retrieval and Recommendation
   Yupeng Hou, Jiacheng Li, Zhankui He, An Yan, Xiusi Chen, Julian McAuley
   arXiv https://arxiv.org/pdf/2403.03952

In this example dataset:
    1. title: replaces a user's search text in the NAPCS dataset
    2. main_category: takes the place of the NAICS description in the
       original training dataset
    3. parent_asin: takes the place of the NAPCS code in the original
       training dataset
"""

import sys
import os
import numpy as np
import pandas as pd
import smart_open

smart_open.open = smart_open.smart_open

from gensim.models.doc2vec import Doc2Vec
from gensim import utils
from nltk.stem import SnowballStemmer
from nltk.corpus import stopwords


def load_labels(repo_path: str) -> pd.Series:
    """load_labels function brings the labels from the training corpus of
    product examples into memory so that we can identify the correct
    code for a given search.

    Args:
        repo_path: the location of the repository

    Returns:
        pd.Series: y_labels is the label for each example

    """
    y_labels = np.load(repo_path + '/labels.npy', allow_pickle=True)
    y_labels = pd.Series(y_labels)

    return y_labels



def stemmer(token: str) -> str:
    """stemmer takes a token and stems it using
       a Snowball Stemmer.

    Args:
        token: a single string token

    Returns:
        cleaned_token: the stemmed token
    """
    cleaned_token = SNOW_BALL.stem(token)

    return cleaned_token


def text_cleaner(search: str, naics_title: str) -> list:
    """text_cleaner takes an input search and a main_category and returns
    a list of tokens that have been stemmed. Stopwords
    and numbers have also been removed.

    Args:
        search: what the respondent typed into the search box
        main_category: the main_category for the product you're searching for

    Returns:
        tokens: a list of tokens without numbers and stop words that has been
                stemmed.
    """
    cleaned_tokens = utils.simple_preprocess(str(search) + ' ' + naics_title)
    cleaned_tokens = [
        stemmer(token)
        for token in cleaned_tokens
        if (token not in STOP_WORDS and not token.isdigit())
    ]
    return cleaned_tokens


def get_top(model, cleaned_tokens: list, epochs=300, start_topn=800):
    """
    get_top takes a doc2vec model and a list of tokens, infers the document
    space embedding for those tokens, and returns a matrix of the top closest
    product data examples

    Parameters:
        model: a trained gensim Doc2Vec model
        search_tokens: a list of tokens that have been preprocessed
        epochs: the number of training epochs we should use when inferring the
                new document vectors for our incoming search
        start_topn: the number of closest documents we should find before
                    deduplication
    """
    # infer the document embeddings of our incoming search
    vector = model.infer_vector(cleaned_tokens, epochs=epochs)
    # collect the top closest product documents to our incoming search
    top = np.array(
        model.docvecs.most_similar([vector], topn=start_topn))
    # get the product codes for those closest documents
    product_codes = Y_LABELS.iloc[top[:, 0]].values
    # combine the product codes and similarities
    top = np.hstack((top[:, 1].reshape(-1, 1), product_codes.reshape(-1, 1)))
    # sort the similarities in descending order
    top = top[top[:, 0].argsort()[::-1]]
    # get indices for duplicate product codes codes
    _, indices = np.unique(top[:, 1], return_index=True)
    # remove those duplicate product codes and return the rest
    top = top[np.sort(indices)]
    return top


def predict(in_content: dict) -> dict:
    """
    Predict takes a search provided by SINCT and returns a set of
    up to 10 recommended product codes.

    Parameter:
        in_content: a dictonary of the form
        {'search_string': text, 'main_category': category}
    Return:
        a dictionary of the form
        {'status': 0, 'message': 'success_main_branch',
         'output':
             {'search': search,
             'main_category': category,
             'product_codes': [list of product codes],
             'Score': [list of scores]}
        }
    """
    # confirm that search and naics code are strings
    search = str(in_content['search_string'])
    main_category = str(in_content['main_category'])
    cleaned_tokens = text_cleaner(search, main_category)  # preprocess search
    # get top closest documents in a data frame
    recs = pd.DataFrame(get_top(SINCT_MODEL, cleaned_tokens),
                        columns=['Score', 'product_codes'])
    # score comes as a string so needs to be converted to float
    recs['Score'] = recs['Score'].astype(float)
    # remove duplicates (in case adding a NAPCS code above creates one)
    # then return top 10 results
    recs = recs.drop_duplicates(subset='NAPCS') \
        .head(n=10) \
        .reset_index(drop=True)

    output = {
        'search': search,
        'main_category': main_category,
        'product_codes': recs['product_codes'].tolist(),
        'Score': recs['Score'].tolist()
    }

    search_results = {
        'status': 0,
        'message': 'success_main_branch',
        'output': output
    }

    return search_results


SEVER_DIR = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(SEVER_DIR + "/server/")
import stub
print("starting server stub")
# define path relative to this directory
SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))
STOP_WORDS = set(stopwords.words('english'))
# SnowballStemmer is an improved version of the Porter Stemmer
SNOW_BALL = SnowballStemmer(language='english')
# load Doc2Vec model trained for SINCT
SINCT_MODEL = Doc2Vec.load(SCRIPT_PATH + '/SINCT_model.model')
# raw training corpus and associated NAPCS codes
Y_LABELS = load_labels(SCRIPT_PATH)
stub.start("SINCT", predict)
